use adw::prelude::*;
use rbw::db::{Entry, EntryData, Field, Uri};
use relm4::gtk::glib::timeout_add_seconds_local;
use relm4::prelude::{AsyncComponentParts, SimpleAsyncComponent};
use relm4::{adw, gtk, AsyncComponentSender, RelmWidgetExt};
use relm4_icons::icon_names;
use std::ffi::OsStr;

#[derive(PartialEq, Clone)]
pub enum DetailMode {
    Display,
    Edit,
}

#[derive(Debug)]
pub enum DetailMsg {
    Edit,
    Discard,
    ClipboardCopy(String),
}

#[derive(Clone)]
pub struct DetailCard {
    mode: DetailMode,
    id: String,
    name: String,
    fields: Vec<Field>,
    cardholder_name: Option<String>,
    number: Option<String>,
    brand: Option<String>,
    exp_month: Option<String>,
    exp_year: Option<String>,
    code: Option<String>,
}

#[relm4::component(async, pub)]
impl SimpleAsyncComponent for DetailCard {
    type Init = Entry;
    type Input = ();
    type Output = DetailMsg;

    view! {
        #[root]
        root_box = &gtk::ScrolledWindow {
            set_hexpand:true,

            gtk::Viewport {
                #[wrap(Some)]
                set_child = &gtk::Box {
                    set_orientation: gtk::Orientation::Vertical,
                    set_hexpand:true,

                    gtk::Label {
                        set_text: model.name.as_ref()
                    },
                    adw::EntryRow {
                        set_editable: model.mode == DetailMode::Edit,
                        set_title: "Cardholder Name",
                        set_text: model.cardholder_name.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy Username",
                            connect_clicked[sender, model] => move |_| {
                                sender.output(DetailMsg::ClipboardCopy(model.cardholder_name.clone().unwrap_or_else(String::new))).unwrap();
                            }
                        }
                    },
                    adw::PasswordEntryRow {
                        set_editable: model.mode == DetailMode::Edit,
                        set_title: "Card Number",
                        set_text: model.number.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy Username",
                            connect_clicked[sender, model] => move |_| {
                                sender.output(DetailMsg::ClipboardCopy(model.number.clone().unwrap_or_else(String::new))).unwrap();
                            }
                        }
                    },
                    adw::PasswordEntryRow {
                        set_editable: model.mode == DetailMode::Edit,
                        set_title: "Security Code",
                        set_text: model.code.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy Username",
                            connect_clicked[sender, model] => move |_| {
                                sender.output(DetailMsg::ClipboardCopy(model.code.clone().unwrap_or_else(String::new))).unwrap();
                            }
                        }
                    },
                    // TODO add icons?
                    adw::EntryRow {
                        set_editable: model.mode == DetailMode::Edit,
                        set_title: "Brand",
                        set_text: model.brand.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy Username",
                            connect_clicked[sender, model] => move |_| {
                                sender.output(DetailMsg::ClipboardCopy(model.brand.clone().unwrap_or_else(String::new))).unwrap();
                            }
                        }
                    },
                    // TODO format it?
                    adw::EntryRow {
                        set_editable: model.mode == DetailMode::Edit,
                        set_title: "Month of Expiration",
                        set_text: model.exp_month.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy Username",
                            connect_clicked[sender, model] => move |_| {
                                sender.output(DetailMsg::ClipboardCopy(model.exp_month.clone().unwrap_or_else(String::new))).unwrap();
                            }
                        }
                    },
                    adw::EntryRow {
                        set_editable: model.mode == DetailMode::Edit,
                        set_title: "Year of Expiration",
                        set_text: model.exp_year.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy Username",
                            connect_clicked[sender, model] => move |_| {
                                sender.output(DetailMsg::ClipboardCopy(model.exp_year.clone().unwrap_or_else(String::new))).unwrap();
                            }
                        }
                    },
                    #[name = "fields_anchor"]
                    gtk::Box {
                        set_orientation: gtk::Orientation::Vertical,
                        set_hexpand: true,
                    },
                }
            }
        }
    }

    async fn init(
        entry_data: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = match entry_data.data {
            EntryData::Login { .. } => todo!(),
            EntryData::Card {
                cardholder_name,
                number,
                brand,
                exp_month,
                exp_year,
                code,
            } => DetailCard {
                mode: DetailMode::Display,
                id: entry_data.id,
                name: entry_data.name,
                fields: entry_data.fields,
                cardholder_name,
                number,
                brand,
                exp_month,
                exp_year,
                code,
            },
            EntryData::Identity { .. } => todo!(),
            EntryData::SecureNote => todo!(),
        };
        let widgets = view_output!();

        for field in model.fields.iter() {
            let field_row = adw::EntryRow::builder()
                .title(field.name.clone().unwrap_or_default())
                .text(field.value.clone().unwrap_or_default())
                .tooltip_text(field.name.clone().unwrap_or_default())
                .editable(model.mode == DetailMode::Edit)
                .build();

            widgets.fields_anchor.append(&field_row);
        }

        AsyncComponentParts { model, widgets }
    }

    async fn update(&mut self, _message: Self::Input, _sender: AsyncComponentSender<Self>) {}
}

#[derive(Clone)]
pub struct DetailLogin {
    mode: DetailMode,
    id: String,
    name: String,
    fields: Vec<Field>,
    username: Option<String>,
    password: Option<String>,
    totp: Option<String>,
    uris: Vec<Uri>,
}

#[relm4::component(async, pub)]
impl SimpleAsyncComponent for DetailLogin {
    type Init = Entry;
    type Input = DetailMsg;
    type Output = DetailMsg;

    view! {
        #[root]
        root_box = &gtk::ScrolledWindow {
            set_hexpand:true,

            gtk::Viewport {

                #[wrap(Some)]
                set_child = &gtk::Box {
                    set_orientation: gtk::Orientation::Vertical,
                    set_hexpand:true,

                    gtk::Label {
                        set_text: model.name.as_ref()
                    },
                    adw::EntryRow {
                        set_title: "Username",
                        set_editable: model.mode == DetailMode::Edit,
                        set_text: model.username.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy Username",
                            connect_clicked[sender, model] => move |_| {
                                sender.output(DetailMsg::ClipboardCopy(model.username.clone().unwrap_or_else(String::new))).unwrap();
                            }
                        }
                    },
                    adw::PasswordEntryRow {
                        set_title: "Password",
                        set_editable: model.mode == DetailMode::Edit,
                        set_text: model.password.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy Password",
                            connect_clicked[sender, model] => move |_| {
                                sender.output(DetailMsg::ClipboardCopy(model.password.clone().unwrap_or_else(String::new))).unwrap();
                            }
                        }
                    },
                    /* TODO if we want to mimic the BW desktop client, we should only
                    *  have 1 EntryRow per TOTP and have it display the key in DisplayMode
                    *  and the URL in EditMode
                    */
                    adw::PasswordEntryRow {
                        set_title: "TOTP URL",
                        set_editable: model.mode == DetailMode::Edit,
                        set_visible: model.totp.as_ref().is_some_and(|value| !value.is_empty()),
                        set_text: model.totp.as_ref().unwrap_or(&String::new()),
                        add_suffix = &gtk::Button {
                            set_icon_name: icon_names::COPY_REGULAR,
                            set_tooltip: "Copy TOTP URL"
                        }
                    },
                    #[name = "totp_anchor"]
                    adw::EntryRow {
                        set_title: "TOTP Code",
                        set_editable: model.mode == DetailMode::Edit,
                        set_text: "0",
                        set_visible: model.totp.as_ref().is_some_and(|value| !value.is_empty()),
                    },
                    #[name = "uris_anchor"]
                    gtk::Box {
                        set_orientation: gtk::Orientation::Vertical,
                        set_hexpand: true,
                    },
                    #[name = "fields_anchor"]
                    gtk::Box {
                        set_orientation: gtk::Orientation::Vertical,
                        set_hexpand: true,
                    },
                    // gtk::Button {
                    //     set_label: "Edit",
                    //     set_visible: model.mode == DetailMode::Display,
                    //     set_icon_name: icon_names::EDIT,
                    //     connect_clicked[sender] => move |_| {
                    //         sender.output(DetailMsg::Edit).unwrap();
                    //     }
                    // }
                }
            }
        }
    }

    async fn init(
        entry_data: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = match entry_data.data {
            EntryData::Login {
                username,
                password,
                totp,
                uris,
            } => DetailLogin {
                mode: DetailMode::Display,
                id: entry_data.id,
                name: entry_data.name,
                fields: entry_data.fields,
                username,
                password,
                totp,
                uris,
            },
            EntryData::Card { .. } => todo!(),
            EntryData::Identity { .. } => todo!(),
            EntryData::SecureNote => todo!(),
        };

        let widgets = view_output!();

        let totp_anchor = widgets.totp_anchor.clone();
        let totp_countdown = gtk::Label::builder().label("0").build();
        totp_anchor.add_suffix(&totp_countdown);
        let copy_totp_button = gtk::Button::builder()
            .icon_name(icon_names::COPY_REGULAR)
            .tooltip_text("Copy TOTP Token")
            .build();
        let sender_clone = sender.clone();
        let totp_anchor_clone = totp_anchor.clone();

        copy_totp_button.connect_clicked(move |_| {
            let string_to_copy = totp_anchor_clone.text();
            sender_clone
                .output(DetailMsg::ClipboardCopy(string_to_copy.to_string()))
                .unwrap();
        });
        totp_anchor.add_suffix(&copy_totp_button);

        let model_clone = model.clone();

        let update_totp = move || {
            if let Some(secret) = model_clone.totp.clone() {
                let secret_str = if let Ok(u) = url::Url::parse(secret.as_str()) {
                    if u.scheme() != "otpauth" {
                        // TODO error handling
                    }
                    if u.host_str() != Some("totp") {
                        // TODO error handling
                    }
                    let query: std::collections::HashMap<_, _> = u.query_pairs().collect();
                    // TODO error handling
                    query.get("secret").unwrap().to_string()
                } else {
                    secret.to_string()
                };
                let totp = base32::decode(
                    base32::Alphabet::RFC4648 { padding: false },
                    // TODO error handling
                    &secret_str.replace(' ', ""),
                )
                .unwrap();

                let totp_key = totp_lite::totp_custom::<totp_lite::Sha1>(
                    totp_lite::DEFAULT_STEP,
                    6,
                    &totp,
                    std::time::SystemTime::now()
                        // TODO error handling
                        .duration_since(std::time::SystemTime::UNIX_EPOCH)
                        .unwrap()
                        .as_secs(),
                );
                let seconds = 30
                    - (std::time::SystemTime::now()
                        // TODO error handling
                        .duration_since(std::time::SystemTime::UNIX_EPOCH)
                        .unwrap()
                        .as_secs()
                        % 30);

                totp_anchor.set_text(&totp_key.as_str());
                totp_countdown.set_text(&seconds.to_string().as_str());
            }
            crate::glib::ControlFlow::Continue
        };

        timeout_add_seconds_local(1, update_totp);

        for uri in model.uris.iter() {
            let copy_button = gtk::Button::builder()
                .icon_name(icon_names::COPY_REGULAR)
                .build();

            let sender_clone = sender.clone();
            let uri_clone = uri.clone();
            copy_button.connect_clicked(move |_| {
                let string_to_copy = uri_clone.uri.clone();
                sender_clone
                    .output(DetailMsg::ClipboardCopy(string_to_copy))
                    .unwrap();
            });

            let open_link_button = gtk::Button::builder()
                .icon_name(icon_names::OPEN_REGULAR)
                .build();

            let uri_clone = uri.clone();
            open_link_button.connect_clicked(move |_| {
                let url = OsStr::new(&uri_clone.uri);
                let _ = open::that(url);
            });

            let uri_row = adw::EntryRow::builder()
                .title("Uri")
                .text(uri.uri.clone())
                .hexpand(true)
                .editable(model.mode == DetailMode::Edit)
                .build();
            uri_row.add_suffix(&open_link_button);
            uri_row.add_suffix(&copy_button);
            widgets.uris_anchor.append(&uri_row);
        }

        for field in model.fields.iter() {
            let field_row = adw::EntryRow::builder()
                .title(field.name.clone().unwrap_or_default())
                .text(field.value.clone().unwrap_or_default())
                .tooltip_text(field.name.clone().unwrap_or_default())
                .editable(model.mode == DetailMode::Edit)
                .build();

            widgets.fields_anchor.append(&field_row);
        }

        AsyncComponentParts { model, widgets }
    }

    async fn update(&mut self, message: Self::Input, _sender: AsyncComponentSender<Self>) {
        match message {
            DetailMsg::Edit => {
                self.mode = DetailMode::Edit;
            }
            DetailMsg::Discard => self.mode = DetailMode::Display,
            // We do not need to do anything here because it will be
            // escalated to the parent component
            DetailMsg::ClipboardCopy(_) => {}
        }
    }
}

#[derive(Clone)]
pub struct DetailIdentity {
    mode: DetailMode,
    id: String,
    name: String,
    fields: Vec<Field>,
    title: Option<String>,
    first_name: Option<String>,
    middle_name: Option<String>,
    last_name: Option<String>,
    address1: Option<String>,
    address2: Option<String>,
    address3: Option<String>,
    city: Option<String>,
    state: Option<String>,
    postal_code: Option<String>,
    country: Option<String>,
    phone: Option<String>,
    email: Option<String>,
    ssn: Option<String>,
    license_number: Option<String>,
    passport_number: Option<String>,
    username: Option<String>,
}

#[relm4::component(async, pub)]
impl SimpleAsyncComponent for DetailIdentity {
    type Init = Entry;
    type Input = DetailMsg;
    type Output = DetailMsg;

    view! {
        #[root]
        root_box = &gtk::ScrolledWindow {
            set_vexpand: true,
            set_hexpand: true,

            gtk::Box {
                set_orientation: gtk::Orientation::Vertical,

                gtk::Label {
                    set_text: model.name.as_ref()
                },
                adw::EntryRow {
                    set_editable: model.mode == DetailMode::Edit,
                    set_visible: model.mode == DetailMode::Edit || (model.mode == DetailMode::Display && model.title.clone().is_some_and(|value| value != "".to_string())),
                    set_title: "Title",
                    set_text: model.title.as_ref().unwrap_or(&String::new()),
                    add_suffix = &gtk::Button {
                        set_icon_name: icon_names::COPY_REGULAR,
                        set_tooltip: "Copy Title",
                        connect_clicked[sender, model] => move |_| {
                            sender.output(DetailMsg::ClipboardCopy(model.title.clone().unwrap_or_else(String::new))).unwrap();
                        }
                    }
                },
                adw::EntryRow {
                    set_editable: model.mode == DetailMode::Edit,
                    set_visible: model.mode == DetailMode::Edit || (model.mode == DetailMode::Display && model.first_name.clone().is_some_and(|value| value != "".to_string())),
                    set_title: "First Name",
                    set_text: model.first_name.as_ref().unwrap_or(&String::new()),
                    add_suffix = &gtk::Button {
                        set_icon_name: icon_names::COPY_REGULAR,
                        set_tooltip: "Copy First Name",
                        connect_clicked[sender, model] => move |_| {
                            sender.output(DetailMsg::ClipboardCopy(model.first_name.clone().unwrap_or_else(String::new))).unwrap();
                        }
                    }
                },
                adw::EntryRow {
                    set_editable: model.mode == DetailMode::Edit,
                    set_visible: model.mode == DetailMode::Edit || (model.mode == DetailMode::Display && model.middle_name.clone().is_some_and(|value| value != "".to_string())),
                    set_title: "Middle Name",
                    set_text: model.middle_name.as_ref().unwrap_or(&String::new()),
                    add_suffix = &gtk::Button {
                        set_icon_name: icon_names::COPY_REGULAR,
                        set_tooltip: "Copy Middle Name",
                        connect_clicked[sender, model] => move |_| {
                            sender.output(DetailMsg::ClipboardCopy(model.middle_name.clone().unwrap_or_else(String::new))).unwrap();
                        }
                    }
                },
                adw::EntryRow {
                    set_editable: model.mode == DetailMode::Edit,
                    set_visible: model.mode == DetailMode::Edit || (model.mode == DetailMode::Display && model.last_name.clone().is_some_and(|value| value != "".to_string())),
                    set_title: "Last Name",
                    set_text: model.last_name.as_ref().unwrap_or(&String::new()),
                    add_suffix = &gtk::Button {
                        set_icon_name: icon_names::COPY_REGULAR,
                        set_tooltip: "Copy Last Name",
                        connect_clicked[sender, model] => move |_| {
                            sender.output(DetailMsg::ClipboardCopy(model.last_name.clone().unwrap_or_else(String::new))).unwrap();
                        }
                    }
                },
                #[name = "fields_anchor"]
                gtk::Box {
                    set_orientation: gtk::Orientation::Vertical,
                    set_hexpand: true,
                }
            }
        }
    }

    async fn init(
        entry_data: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = match entry_data.data {
            EntryData::Login { .. } => todo!(),
            EntryData::Card { .. } => todo!(),
            EntryData::Identity {
                title,
                first_name,
                middle_name,
                last_name,
                address1,
                address2,
                address3,
                city,
                state,
                postal_code,
                country,
                phone,
                email,
                ssn,
                license_number,
                passport_number,
                username,
            } => DetailIdentity {
                mode: DetailMode::Display,
                id: entry_data.id,
                name: entry_data.name,
                fields: entry_data.fields,
                title,
                first_name,
                middle_name,
                last_name,
                address1,
                address2,
                address3,
                city,
                state,
                postal_code,
                country,
                phone,
                email,
                ssn,
                license_number,
                passport_number,
                username,
            },
            EntryData::SecureNote => todo!(),
        };

        let widgets = view_output!();

        for field in model.fields.iter() {
            let field_row = adw::EntryRow::builder()
                .title(field.name.clone().unwrap_or_default())
                .text(field.value.clone().unwrap_or_default())
                .tooltip_text(field.name.clone().unwrap_or_default())
                .editable(model.mode == DetailMode::Edit)
                .build();

            widgets.fields_anchor.append(&field_row);
        }

        AsyncComponentParts { model, widgets }
    }

    async fn update(&mut self, message: Self::Input, _sender: AsyncComponentSender<Self>) {}
}

#[derive(Clone)]
pub struct DetailNote {
    mode: DetailMode,
    id: String,
    name: String,
    fields: Vec<Field>,
}

#[relm4::component(async, pub)]
impl SimpleAsyncComponent for DetailNote {
    type Init = Entry;
    type Input = DetailMsg;
    type Output = DetailMsg;

    view! {
        #[root]
        root_box = &gtk::ScrolledWindow {
            set_vexpand: true,
            set_hexpand: true,

            gtk::Box {
                set_orientation: gtk::Orientation::Vertical,

                gtk::Label {
                    set_text: model.name.as_ref()
                },
                #[name = "fields_anchor"]
                gtk::Box {
                    set_orientation: gtk::Orientation::Vertical,
                    set_hexpand: true,
                }
            }
        }
    }

    async fn init(
        entry_data: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = match entry_data.data {
            EntryData::Login { .. } => todo!(),
            EntryData::Card { .. } => todo!(),
            EntryData::Identity { .. } => todo!(),
            EntryData::SecureNote => DetailNote {
                mode: DetailMode::Display,
                id: entry_data.id,
                name: entry_data.name,
                fields: entry_data.fields,
            },
        };

        let widgets = view_output!();

        for field in model.fields.iter() {
            let field_row = adw::EntryRow::builder()
                .title(field.name.clone().unwrap_or_default())
                .text(field.value.clone().unwrap_or_default())
                .tooltip_text(field.name.clone().unwrap_or_default())
                .editable(model.mode == DetailMode::Edit)
                .build();

            widgets.fields_anchor.append(&field_row);
        }

        AsyncComponentParts { model, widgets }
    }

    async fn update(&mut self, message: Self::Input, _sender: AsyncComponentSender<Self>) {}
}
