use std::time::Duration;

use adw::prelude::*;
use rbw::db::{Entry, EntryData};
use relm4::factory::AsyncFactoryComponent;
use relm4::prelude::DynamicIndex;
use relm4::{adw, gtk, AsyncFactorySender};
use relm4_icons::icon_names;

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone)]
pub enum VaultItemType {
    Login,
    Card,
    Identity,
    Note,
}

#[derive(Clone)]
pub struct ListItem {
    pub id: String,
    pub primary: String,
    pub secondary: String,
    pub item_type: VaultItemType,
}

#[derive(Debug)]
pub enum ListOutput {
    Select(String),
}

#[relm4::factory(async, pub)]
impl AsyncFactoryComponent for ListItem {
    type Init = ListItem;
    type Input = ();
    type Output = ListOutput;
    type CommandOutput = ();
    type ParentWidget = gtk::ListBox;

    view! {
        root =
            adw::ActionRow {
                set_title: self.primary.as_str(),
                set_subtitle: self.secondary.as_str(),
                set_activatable: true,
                set_selectable: true,
                add_prefix = &gtk::Image {
                    set_icon_name: Some( match self.item_type {
                        VaultItemType::Login => icon_names::PERMISSIONS_PASSWORD,
                        VaultItemType::Card => icon_names::VCARD,
                        VaultItemType::Identity => icon_names::PERSON,
                        VaultItemType::Note => icon_names::NOTE
                    })
                },
                connect_activated[sender, model = self.clone()] => move |_| {
                    let _ = sender.output(ListOutput::Select(model.id.clone()));
                },
                set_hexpand: true,
            }
    }

    async fn init_model(
        item: Self::Init,
        _index: &DynamicIndex,
        _sender: AsyncFactorySender<Self>,
    ) -> Self {
        tokio::time::sleep(Duration::from_secs(1)).await;
        Self {
            id: item.id,
            primary: item.primary,
            secondary: item.secondary,
            item_type: item.item_type,
        }
    }

    async fn update(&mut self, _msg: Self::Input, _sender: AsyncFactorySender<Self>) {
        tokio::time::sleep(Duration::from_secs(1)).await;
    }
}

impl From<Entry> for ListItem {
    fn from(entry: Entry) -> Self {
        ListItem {
            id: entry.id,
            item_type: match entry.data {
                EntryData::Login { .. } => VaultItemType::Login,
                EntryData::Card { .. } => VaultItemType::Card,
                EntryData::Identity { .. } => VaultItemType::Identity,
                EntryData::SecureNote => VaultItemType::Note,
            },
            primary: if !entry.name.is_empty() {
                entry.name
            } else {
                String::new()
            },
            secondary: match entry.data {
                EntryData::Login { username, .. } => {
                    if username.clone().is_some_and(|n| !n.is_empty()) {
                        username.unwrap()
                    } else {
                        String::new()
                    }
                }
                EntryData::Card {
                    cardholder_name,
                    .. /*
                    number,
                    brand,
                    exp_month,
                    exp_year,
                    code, */
                } => {
                    if cardholder_name.clone().is_some_and(|n| !n.is_empty()) {
                        cardholder_name.unwrap()
                    } else {
                        String::new()
                    }
                }
                // TODO correct subtitle pls
                EntryData::Identity {
                    .. /*
                    title,
                    first_name,
                    middle_name,
                    last_name,
                    address1,
                    address2,
                    address3,
                    city,
                    state,
                    postal_code,
                    country,
                    phone,
                    email,
                    ssn,
                    license_number,
                    passport_number,
                    username, */
                } => String::new(),
                EntryData::SecureNote => String::new(),
            },
        }
    }
}
