use std::str::FromStr;

use rbw::{
    db::{Entry, EntryData},
    error::Error,
    locked::Keys,
};
use sqlx::{
    sqlite::{SqliteConnectOptions, SqliteJournalMode},
    SqlitePool,
};

use crate::{components::vault::VaultCreationSetting, Vault, VaultSetting};

pub fn decrypt(mut entry: Entry, unlocked_keys: &Keys) -> Result<Entry, Error> {
    // TODO implement decryption for each field
    // Missing fields SecureNote
    entry.name = decrypt_string(&entry.name, unlocked_keys);
    entry.fields = entry
        .fields
        .iter_mut()
        .map(|field| {
            field.name = if let Some(name) = field.name.clone() {
                Some(decrypt_string(name.as_str(), unlocked_keys))
            } else {
                None
            };
            field.value = if let Some(value) = field.value.clone() {
                Some(decrypt_string(value.as_str(), unlocked_keys))
            } else {
                None
            };
            field.to_owned()
        })
        .collect();
    entry.data = match entry.data {
        EntryData::Login {
            username,
            password,
            totp,
            uris,
        } => EntryData::Login {
            username: Some(decrypt_string(
                username.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            password: Some(decrypt_string(
                password.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            totp: Some(decrypt_string(
                totp.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            uris: uris
                .clone()
                .iter_mut()
                .map(|uri| {
                    uri.uri = decrypt_string(uri.uri.as_str(), unlocked_keys);
                    uri.to_owned()
                })
                .collect(),
        },
        EntryData::Card {
            cardholder_name,
            number,
            brand,
            exp_month,
            exp_year,
            code,
        } => EntryData::Card {
            cardholder_name: Some(decrypt_string(
                cardholder_name.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            number: Some(decrypt_string(
                number.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            brand: Some(decrypt_string(
                brand.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            exp_month: Some(decrypt_string(
                exp_month.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            exp_year: Some(decrypt_string(
                exp_year.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            code: Some(decrypt_string(
                code.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
        },
        EntryData::Identity {
            title,
            first_name,
            middle_name,
            last_name,
            address1,
            address2,
            address3,
            city,
            state,
            postal_code,
            country,
            phone,
            email,
            ssn,
            license_number,
            passport_number,
            username,
        } => EntryData::Identity {
            title: Some(decrypt_string(
                title.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            first_name: Some(decrypt_string(
                first_name.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            middle_name: Some(decrypt_string(
                middle_name.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            last_name: Some(decrypt_string(
                last_name.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            address1: Some(decrypt_string(
                address1.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            address2: Some(decrypt_string(
                address2.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            address3: Some(decrypt_string(
                address3.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            city: Some(decrypt_string(
                city.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            state: Some(decrypt_string(
                state.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            postal_code: Some(decrypt_string(
                postal_code.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            country: Some(decrypt_string(
                country.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            phone: Some(decrypt_string(
                phone.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            email: Some(decrypt_string(
                email.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            ssn: Some(decrypt_string(
                ssn.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            license_number: Some(decrypt_string(
                license_number.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            passport_number: Some(decrypt_string(
                passport_number.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            username: Some(decrypt_string(
                username.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
        },
        EntryData::SecureNote => EntryData::SecureNote,
    };
    Ok(entry)
}

pub fn decrypt_string(string_to_decrypt: &str, unlocked_keys: &Keys) -> String {
    if !string_to_decrypt.is_empty() {
        let x: &[_] = &[
            '\t', '\n', '\r', '\u{1}', '\u{2}', '\u{3}', '\u{4}', '\u{5}', '\u{6}', '\u{7}',
            '\u{8}', '\u{10}', '\u{b}', '\u{c}', '\u{e}', '\u{f}',
        ];
        let cipher_id: rbw::cipherstring::CipherString =
            rbw::cipherstring::CipherString::new(string_to_decrypt).unwrap();
        let decrypted_id = cipher_id
            .decrypt_locked_symmetric(unlocked_keys)
            .inspect_err(|e| eprintln!("failed decrypting id: {e}"));
        std::str::from_utf8(decrypted_id.unwrap().data())
            .unwrap()
            .trim_matches(x)
            .to_string()
    } else {
        String::new()
    }
}

pub async fn init_db() -> Result<(), sqlx::Error> {
    let database_url =
        std::env::var("DATABASE_URL").unwrap_or("sqlite://vault_settings.db3".to_string());
    let options = SqliteConnectOptions::from_str(&database_url)
        .expect("Error parsing database URL")
        .create_if_missing(true)
        .journal_mode(SqliteJournalMode::Wal);

    let pool = SqlitePool::connect_with(options)
        .await
        .expect("Error connecting to database");

    sqlx::migrate!()
        .run(&pool)
        .await
        .expect("Error when applying database migrations");
    Ok(())
}

struct ActiveVault {
    base_url: String,
    email: String,
}

pub async fn load_active_vault_from_db() -> Result<Option<Vault>, sqlx::Error> {
    let pool = SqlitePool::connect("sqlite://vault_settings.db3").await?;

    let active_vault_setting =
        sqlx::query_as!(ActiveVault, "SELECT base_url, email FROM active_vault")
            .fetch_one(&pool)
            .await?;

    let active_vault = sqlx::query_as!(
        VaultSetting,
        "SELECT base_url, email, identity_url 
            FROM vault_settings 
            WHERE base_url = ?
              AND email = ?",
        active_vault_setting.base_url,
        active_vault_setting.email
    )
    .fetch_one(&pool)
    .await?;

    Ok(Some(Vault::from(active_vault)))
}

pub async fn save_active_vault_to_db(setting: VaultSetting) -> Result<(), sqlx::Error> {
    let pool = SqlitePool::connect("sqlite://vault_settings.db3").await?;

    if let Ok(_existing_row) = sqlx::query_as!(
        ActiveVault,
        "SELECT base_url, email
            FROM active_vault"
    )
    .fetch_one(&pool)
    .await
    {
        let _update_active_result = sqlx::query!(
            "UPDATE active_vault
                    SET (base_url, email)
                    = (?1, ?2)",
            setting.base_url,
            setting.email
        )
        .execute(&pool)
        .await?;
    } else {
        let _insert_result = sqlx::query!(
            "INSERT INTO active_vault (base_url, email)
            VALUES (?1, ?2)",
            setting.base_url,
            setting.email,
        )
        .execute(&pool)
        .await?;
    }

    Ok(())
}

pub async fn load_vault_settings_from_db() -> Result<Vec<VaultSetting>, sqlx::Error> {
    let pool = SqlitePool::connect("sqlite://vault_settings.db3").await?;

    let account_settings = sqlx::query_as!(
        VaultSetting,
        "SELECT base_url, email, identity_url FROM vault_settings"
    )
    .fetch_all(&pool)
    .await?;

    Ok(account_settings)
}

pub async fn insert_vault_setting_to_db(setting: VaultCreationSetting) -> Result<(), sqlx::Error> {
    let pool = SqlitePool::connect("sqlite://vault_settings.db3").await?;

    let identity_url = setting.identity_url.clone().unwrap_or_default();

    // TODO maybe we want a transaction around this to roll back?
    let _insert_result = sqlx::query!(
        "INSERT INTO vault_settings (base_url, email, identity_url)
            VALUES (?1, ?2, ?3)",
        setting.base_url,
        setting.email,
        identity_url
    )
    .execute(&pool)
    .await?;

    if setting.set_active {
        let _save_result = save_active_vault_to_db(VaultSetting::from(setting)).await;
    }

    Ok(())
}
